/*
* by
* Massimiliano Corsini,
* Visual Computing Lab
* ISTI-CNR, Pisa, Italy
*/

// Local headers
#include "glwidget.h"

// Standard headers
#include <math.h>
#include <iostream>

// Qt headers
#include <QMouseEvent>

// inclusion of the nvmc files
#include "client_game.h"	// all the concepts we need

// VCG headers
#include <vcg/space/point3.h>

/** Client  */
extern  NvmcClient cli;

/* value of pi */
#define M_PI       3.14159265358979323846

/* send a snapshot */
bool sendSS = false;


using namespace vcg;

GLWidget::GLWidget(QWidget *parent/*=NULL*/): QGLWidget(parent)
{
    QGLFormat qFormat = QGLFormat::defaultFormat();
    qFormat.setAlpha(true);
    qFormat.setDepth(true);
    qFormat.setDoubleBuffer(true);
    setFormat(qFormat);

    setAutoBufferSwap(false);
    setFocusPolicy(Qt::ClickFocus);

    tm.start();
}

void GLWidget::initializeGL()
{
    // set viewport
    glViewport(0, 0, width(), height());

    // init GLEW
    GLenum err = glewInit();
    if (GLEW_OK != err)
    {
        std::cout << "GLEW initialization error!" << " (" << glewGetErrorString(err) << ")" << endl;
        exit(-1);
    }

    // in questo modo soltanto i contorni dei poligoni vengono rasterizzati
    glPolygonMode(GL_FRONT,GL_FILL);

    std::cout << "OpenGL initialized" << endl;
}

void GLWidget::paintGL()
{
    drawScene();
    swapBuffers();
}

void GLWidget::drawScene()
{
    static int cn = 0;
    static int part = 0;
    glClear(GL_COLOR_BUFFER_BIT);

    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();

    // here we take the bounding box of the whole circuit
    vcg::Point3f bmin = race.circuit->bbox.min;
    vcg::Point3f bmax = race.circuit->bbox.max;

    float sidex,sidez;
    sidex = bmax[0]-bmin[0];
    sidez = bmax[2]-bmin[2];

    // this function asks nvmc to update the state of the player with the information
    // received from the server
    cli.UpdatePlayers();

    // set the camera

    Race::PlayersIterator  hh  = cli.local_player;

    if( hh != cli.players.end())
    {
        glOrtho(-6,6,-6,6,0.0,10.0);
        glMatrixMode(GL_MODELVIEW);
        glLoadIdentity();

        vcg::Point4f pos = (*hh).car.frame.O();
        gluLookAt(pos[0],6.0,pos[2],pos[0],pos[1],pos[2],0,0,-1);
    }
    else// if there are no player just render the circuit from above
    {
        glOrtho(-sidex*0.5,sidex*0.5,-sidez*0.5,sidez*0.5,0.0,400);
        glMatrixMode(GL_MODELVIEW);
        glLoadIdentity();
        gluLookAt(0,400,0,0,0,0,0,0,-1);
    }

    // draw bounding box of the terrain
    glColor3f(0.1f, 0.2f, 0.4f);
    glBegin(GL_QUADS);
    glVertex3f(bmin[0],bmin[1],bmin[2]);
    glVertex3f(bmax[0],bmin[1],bmin[2]);
    glVertex3f(bmax[0],bmin[1],bmax[2]);
    glVertex3f(bmin[0],bmin[1],bmax[2]);
    glEnd();

    // draw the axis
    glPushMatrix();
    glScalef(bmax[0]-bmin[0],bmax[1]-bmin[1],bmax[2]-bmin[2]);
    draw_axis();
    glPopMatrix();

    // draw the circuit
    Street::TilesIterator ti;
    int i;
    glColor3f(1.0,0.0,0.0);
    glPolygonMode(GL_FRONT,GL_LINE);
    glBegin(GL_QUAD_STRIP);
    for(i=0,ti = race.circuit->street.path.begin(); ti != race.circuit->street.path.end();++ti,++i)
    {
        vcg::Point3f p0,p1;
        p0 = (*ti)[0].P(); p1 = (*ti)[1].P();
        glVertex3f(p0[0],p0[1]+0.1,p0[2]);
        glVertex3f(p1[0],p1[1]+0.1,p1[2]);
    }
    glEnd();

    // draw the tunnels
    Street::TunnelsIterator  tni;
    glColor3f(1.0,1.0,0.0);
    glPolygonMode(GL_FRONT,GL_FILL);
    for(i=0,tni = race.circuit->street.tunnels.begin(); tni != race.circuit->street.tunnels.end();++tni,++i)
    {
        glBegin(GL_QUAD_STRIP);
        for(i=0,ti = (*tni).path.begin(); ti != (*tni).path.end();++ti,++i)
        {
            vcg::Point3f p0,p1;
            p0 = (*ti)[0].P(); p1 = (*ti)[1].P();
            glVertex3f(p0[0],p0[1]+0.2,p0[2]);
            glVertex3f(p1[0],p1[1]+0.2,p1[2]);
        }
        glEnd();
    }

    // draw the buildings
    Circuit::BuildingsIterator bi;
    glColor3f(1.0,1.0,1.0);
    glPolygonMode(GL_FRONT,GL_FILL);
    for(bi = race.circuit->buildings.begin(); bi != race.circuit->buildings.end(); ++bi)
    {
        glBegin(GL_POLYGON);

        for(unsigned i=0; i < (*bi).footprint.size();++i)
            glVertex3f((*bi).footprint[i].P()[0],(*bi).footprint[i].P()[1],(*bi).footprint[i].P()[2]);

        glEnd();
    }

    // draw the trees
    Circuit::TreesIterator tsi;
    glColor3f(10.0,1.0,0.0);
    glBegin(GL_POINTS);

    for(tsi = race.circuit->trees.begin(); tsi != race.circuit->trees.end(); ++tsi)
              glVertex3f((*tsi).pos[0],(*tsi).pos[1],(*tsi).pos[2]);

    glEnd();

    // draw the players
    Race::PlayersIterator pi;

    for(pi = cli.players.begin(); pi != cli.players.end(); ++pi)
    {
        glPushMatrix();

        //apply transformation coded in the frame
        Frame f = ((*pi).car.frame);
        glMultMatrixf(&f[0][0]);
        glPushMatrix();
        glColor3f(1.0,1.0,1.0);

        glColor3f(0,1,0);
        glBegin(GL_LINES);
        glVertex3f(0.0,1.0,1.0);
        glVertex3f(0.0,1.0,-1.0);
        glEnd();

        glColor3f(0,0,1);
        glTranslatef(0.0,0.0,(*pi).car.front_wheel_pos);
        glRotatef((*pi).car.steering_angle*180.0/ M_PI ,0,1,0);
        glBegin(GL_LINES);
        glVertex3f(0.0f, 1.0f, 0.4f);
        glVertex3f(0.0f, 1.0f, -0.4f);
        glEnd();

        glPopMatrix();
        glColor3f(1.0,0.0,0.0);
        renderText(0.0,0.0,QString((*pi).ip_address.toString())); // write the IP address of the client
        glPopMatrix();
    }

    glLoadIdentity();

    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();

    // print the commands
    glColor3f(1.0,1.0,0.0);

    QFont serifFont("Arial", 12, QFont::Bold);
    renderText(20,20,QString("accel: w"),serifFont);
    renderText(20,40,QString("slow:  s"),serifFont);
    renderText(20,60,QString("left:  a"),serifFont);
    renderText(20,80,QString("right: d"),serifFont);
    renderText(20,100,QString("brake: x"),serifFont);
    renderText(20,120,QString("reset: p"),serifFont);

    if(sendSS)
    {
        cli.SendScreenshot();
        sendSS=false;
    }

    cli.DrawScreenshots(std::string());

    cn++;
    if (tm.elapsed() > 1000)
    {
        std::cout << "VIEW fps " << (1000.0*cn)/tm.elapsed() << std::endl;
        tm.restart();
        cn=0;
    }
}

void GLWidget::resizeGL(int width, int height)
{
    glViewport(0,0,width,height);

}

void GLWidget::mousePressEvent(QMouseEvent *event)
{
    // mouse press
    // not handled
}

void GLWidget::mouseMoveEvent(QMouseEvent *event)
{
    // mouse move
    // not handled
}

void GLWidget::mouseReleaseEvent(QMouseEvent *event)
{
    // mouse release
    // not handled
}

void GLWidget::wheelEvent(QWheelEvent *we)
{
    // mouse wheel
    // not handled
}

void GLWidget::keyPressEvent(QKeyEvent *ke)
{
    int key = ke->nativeVirtualKey();

    if (key == Qt::Key_W)
    {
        // increase speed
        cli.Command(COMMAND_INCREASE_SPEED);
    }

    else if (key == Qt::Key_S)
    {
        // decrease speed
        cli.Command(COMMAND_DECREASE_SPEED);
    }
    else if (key == Qt::Key_A)
    {
        // turn left
        cli.Command(COMMAND_TURN_LEFT);
    }
    else if (key == Qt::Key_D)
    {
        // turn right
        cli.Command(COMMAND_TURN_RIGHT);
    }
    else if (key == Qt::Key_X)
    {
        cli.Command(COMMAND_BRAKE);
    }
    else if (key == Qt::Key_P)
    {
        cli.Command(COMMAND_RESTART);
    }
    else if (key == Qt::Key_G)
    {
        sendSS = true;
    }
}

// draw reference system
void GLWidget::draw_axis()
{
    glPushAttrib(0xffffffff);
    glDisable(GL_LIGHTING);

    glBegin(GL_LINES);

    // X
    glColor3f(1.0,0.0,0.0);
    glVertex3f(0,0,0);
    glVertex3f(10,0,0);

    // Y
    glColor3f(0.0,1.0,0.0);
    glVertex3f(0,0,0);
    glVertex3f(0,10,0);

    // Z
    glColor3f(0.0,0.0,1.0);
    glVertex3f(0,0,0);
    glVertex3f(0,0,10);

    glEnd();

    glPopAttrib();
}
