/*
* by
* Massimiliano Corsini,
* Visual Computing Lab
* ISTI-CNR, Pisa, Italy
*/

// Local headers
#include "glwidget.h"

// Standard headers
#include <math.h>
#include <iostream>

// Qt headers
#include <QMouseEvent>

// VCG headers
#include <wrap/gl/space.h>

/* send a snapshot */
bool sendSS = false;

GLWidget::GLWidget(VCGMesh *m, QWidget *parent/*=NULL*/): mesh(m), QGLWidget(parent)
{
    QGLFormat qFormat = QGLFormat::defaultFormat();
    qFormat.setAlpha(true);
    qFormat.setDepth(true);
    qFormat.setDoubleBuffer(true);
    setFormat(qFormat);

    setAutoBufferSwap(false);
    setFocusPolicy(Qt::ClickFocus);

    // 1 degree
    step = 2.5f;

    rx = 0.0f;
    ry = 15.0f;
    rz = 0.0f;
}

void GLWidget::initializeGL()
{
    // set viewport
    glViewport(0, 0, width(), height());

    // init GLEW
    GLenum err = glewInit();
    if (GLEW_OK != err)
    {
        std::cout << "GLEW initialization error!" << " (" << glewGetErrorString(err) << ")" << endl;
        exit(-1);
    }

    glViewport(0,0,width(),height());
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    gluPerspective(40.0, (float)width()/(float)height(), 0.1, 100.0);

    // enable back-face culling
    glEnable(GL_CULL_FACE);

    // enable lighting
    glEnable(GL_LIGHTING);
    glEnable(GL_LIGHT0);

    // enable z-buffer
    glEnable(GL_DEPTH_TEST);

    // normals are re-normalized
    glEnable(GL_NORMALIZE);

    std::cout << "OpenGL initialized" << std::endl;
}

void GLWidget::paintGL()
{
    drawScene();
    swapBuffers();
}

void GLWidget::drawScene()
{
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();


    // lighting and material settings
    ////////////////////////////////////////////////////////////

    GLfloat light_diffuse[] = {1.0f, 1.0f, 1.0f, 1.0f};
    GLfloat light_specular[] = {1.0f, 1.0f, 1.0f, 1.0f};
    GLfloat light_position[] = {1.0f, 1.0f, 1.0f, 1.0f};
    glLightfv(GL_LIGHT0, GL_DIFFUSE, light_diffuse);
    glLightfv(GL_LIGHT0, GL_SPECULAR, light_specular);
    glLightfv(GL_LIGHT0, GL_POSITION, light_position);

    GLfloat mat_ambient[] = {0.0f, 0.0f, 0.0f, 0.0f};
    GLfloat mat_diffuse[] = {0.0f, 0.4f, 1.0f, 1.0f};
    GLfloat mat_specular[] = {1.0f, 1.0f, 1.0f, 1.0f};
    GLfloat mat_shininess[] = {30.0f};
    glMaterialfv(GL_FRONT, GL_AMBIENT, mat_ambient);
    glMaterialfv(GL_FRONT, GL_DIFFUSE, mat_diffuse);
    glMaterialfv(GL_FRONT, GL_SPECULAR, mat_specular);
    glMaterialfv(GL_FRONT, GL_SHININESS, mat_shininess);

    // shading with Gouraud shading
    glShadeModel(GL_SMOOTH);

    glLightf(GL_LIGHT0, GL_CONSTANT_ATTENUATION, 1.0f);
    glLightf(GL_LIGHT0, GL_LINEAR_ATTENUATION, 0.0f);
    glLightf(GL_LIGHT0, GL_QUADRATIC_ATTENUATION, 0.0f);

    // geometric transformation
    ////////////////////////////////////////////////////////////

    // from world-space to view-space
    gluLookAt(0,0,1.5, 0,0,0, 0,1,0);

    // from object-space to world-space
    // rotation+traslation+scaling

    // rotation around X axis
    glRotatef(rx, 1.0f,0.0f,0.0f);

    // rotation around Y axis
    glRotatef(ry, 0.0f,1.0f,0.0f);

    // rotation around Z axis
    glRotatef(rz, 0.0f,0.0f,1.0f);

    // scaling to fix the view frustum
    float d = 1.0f/mesh->bbox.Diag();
    vcg::glScale(d);

    // translate into the origin
    vcg::glTranslate(-mesh->bbox.Center());

    // sending geometry to the graphics hardware
    ////////////////////////////////////////////////////////////

    VCGMesh::VertexPointer v1,v2,v3;
    VCGMesh::FaceIterator itF;

    glBegin(GL_TRIANGLES);

    for (itF = mesh->face.begin(); itF != mesh->face.end(); itF++)
    {
        v1 = (*itF).V(0);
        v2 = (*itF).V(1);
        v3 = (*itF).V(2);

        vcg::glNormal(v1->N());
        vcg::glVertex(v1->P());
        vcg::glNormal(v2->N());
        vcg::glVertex(v2->P());
        vcg::glNormal(v3->N());
        vcg::glVertex(v3->P());
    }

    glEnd();
}

void GLWidget::resizeGL(int width, int height)
{
    glViewport(0,0,width,height);
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    gluPerspective(40.0, (float)width/(float)height, 0.1, 100.0);
}

void GLWidget::mousePressEvent(QMouseEvent *event)
{
    // mouse press
    // not handled
}

void GLWidget::mouseMoveEvent(QMouseEvent *event)
{
    // mouse move
    // not handled
}

void GLWidget::mouseReleaseEvent(QMouseEvent *event)
{
    // mouse release
    // not handled
}

void GLWidget::wheelEvent(QWheelEvent *we)
{
    // mouse wheel
    // not handled
}

void GLWidget::keyPressEvent(QKeyEvent *ke)
{
    int key = ke->nativeVirtualKey();

    if (key == Qt::Key_Q)
    {
        rx -= step;
        repaint();
    }
    else if (key == Qt::Key_W)
    {
        rx += step;
        repaint();
    }
    else if (key == Qt::Key_A)
    {
        ry -= step;
        repaint();
    }
    else if (key == Qt::Key_S)
    {
        ry += step;
        repaint();
    }
    else if (key == Qt::Key_Z)
    {
        rz -= step;
        repaint();
    }
    else if (key == Qt::Key_X)
    {
        rz += step;
        repaint();
    }
}

// draw reference system
void GLWidget::draw_axis()
{
    glPushAttrib(0xffffffff);
    glDisable(GL_LIGHTING);

    glBegin(GL_LINES);

    // X
    glColor3f(1.0,0.0,0.0);
    glVertex3f(0,0,0);
    glVertex3f(10,0,0);

    // Y
    glColor3f(0.0,1.0,0.0);
    glVertex3f(0,0,0);
    glVertex3f(0,10,0);

    // Z
    glColor3f(0.0,0.0,1.0);
    glVertex3f(0,0,0);
    glVertex3f(0,0,10);

    glEnd();

    glPopAttrib();
}
